import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.utils.testing_utils import assert_timeseries_near_equal
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


@use_tempdirs
class TestIntegrateControl(unittest.TestCase):

    def _test_integrate_control(self, transcription):

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta', fix_initial=False, rate_source='theta_rate', targets=['theta'])

        # Define theta as a control.
        phase.add_control(name='theta_rate', units='rad/s', shape=(1,), targets=None)

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        p.set_val('traj.phase0.t_initial', 0.0, units='s')
        p.set_val('traj.phase0.t_duration', 5.0, units='s')

        p.set_val('traj.phase0.states:x',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:y',
                  phase.interpolate(ys=[10, 5], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:v',
                  phase.interpolate(ys=[0, 5], nodes='state_input'),
                  units='m/s')

        p.set_val('traj.phase0.states:int_theta',
                  phase.interpolate(ys=[0.1, 45], nodes='state_input'),
                  units='deg')

        p.set_val('traj.phase0.controls:theta_rate',
                  phase.interpolate(ys=[10, 10], nodes='control_input'),
                  units='deg/s')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=False)

        sol = om.CaseReader('dymos_solution.db').get_case('final')
        sim = om.CaseReader('dymos_simulation.db').get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.states:x')
        x_sim = sim.get_val('traj.phase0.timeseries.states:x')

        y_sol = sol.get_val('traj.phase0.timeseries.states:y')
        y_sim = sim.get_val('traj.phase0.timeseries.states:y')

        v_sol = sol.get_val('traj.phase0.timeseries.states:v')
        v_sim = sim.get_val('traj.phase0.timeseries.states:v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.states:int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.states:int_theta')

        theta_rate_sol = sol.get_val('traj.phase0.timeseries.controls:theta_rate')
        theta_rate_sim = sim.get_val('traj.phase0.timeseries.controls:theta_rate')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, theta_rate_sol, t_sim, theta_rate_sim, tolerance=1.0E-3)

    def _test_integrate_control_rate(self, transcription):

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta', fix_initial=False, rate_source='theta_rate', targets=['theta'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', shape=(1,), targets=None)

        # Force the initial value of the theta polynomial control to equal the initial value of the theta state.
        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='theta', var_b='int_theta',
                                    loc_a='initial', loc_b='initial')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        p.set_val('traj.phase0.t_initial', 0.0, units='s')
        p.set_val('traj.phase0.t_duration', 5.0, units='s')

        p.set_val('traj.phase0.states:x',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:y',
                  phase.interpolate(ys=[10, 5], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:v',
                  phase.interpolate(ys=[0, 5], nodes='state_input'),
                  units='m/s')

        p.set_val('traj.phase0.states:int_theta',
                  phase.interpolate(ys=[0.1, 45], nodes='state_input'),
                  units='deg')

        p.set_val('traj.phase0.controls:theta',
                  phase.interpolate(ys=[0, 100], nodes='control_input'),
                  units='deg')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=False)

        sol = om.CaseReader('dymos_solution.db').get_case('final')
        sim = om.CaseReader('dymos_simulation.db').get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.states:x')
        x_sim = sim.get_val('traj.phase0.timeseries.states:x')

        y_sol = sol.get_val('traj.phase0.timeseries.states:y')
        y_sim = sim.get_val('traj.phase0.timeseries.states:y')

        v_sol = sol.get_val('traj.phase0.timeseries.states:v')
        v_sim = sim.get_val('traj.phase0.timeseries.states:v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.states:int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.states:int_theta')

        theta_sol = sol.get_val('traj.phase0.timeseries.controls:theta')
        theta_sim = sim.get_val('traj.phase0.timeseries.controls:theta')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, tolerance=1.0E-3)

        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sol, theta_sol, tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sim, int_theta_sim, t_sim, theta_sim, tolerance=1.0E-2)

    def _test_integrate_control_rate2(self, transcription):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta_dot', fix_initial=False, rate_source='theta_rate2')
        phase.add_state('int_theta', fix_initial=False, rate_source='int_theta_dot', targets=['theta'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', shape=(1,), targets=None)

        # Force the initial value of the theta polynomial control to equal the initial value of the theta state.
        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='theta', var_b='int_theta',
                                    loc_a='initial', loc_b='initial')

        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='int_theta_dot', var_b='theta_rate',
                                    loc_a='initial', loc_b='initial',
                                    units='rad/s')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.pyOptSparseDriver(optimizer='SLSQP')

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        p.set_val('traj.phase0.t_initial', 0.0, units='s')
        p.set_val('traj.phase0.t_duration', 5.0, units='s')

        p.set_val('traj.phase0.states:x',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:y',
                  phase.interpolate(ys=[10, 5], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:v',
                  phase.interpolate(ys=[0, 5], nodes='state_input'),
                  units='m/s')

        p.set_val('traj.phase0.states:int_theta',
                  phase.interpolate(ys=[0.1, 45], nodes='state_input'),
                  units='deg')

        p.set_val('traj.phase0.states:int_theta_dot',
                  phase.interpolate(ys=[0.0, 0.0], nodes='state_input'),
                  units='deg/s')

        p.set_val('traj.phase0.controls:theta',
                  phase.interpolate(ys=[0, 100], nodes='control_input'),
                  units='deg')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=False)

        sol = om.CaseReader('dymos_solution.db').get_case('final')
        sim = om.CaseReader('dymos_simulation.db').get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.states:x')
        x_sim = sim.get_val('traj.phase0.timeseries.states:x')

        y_sol = sol.get_val('traj.phase0.timeseries.states:y')
        y_sim = sim.get_val('traj.phase0.timeseries.states:y')

        v_sol = sol.get_val('traj.phase0.timeseries.states:v')
        v_sim = sim.get_val('traj.phase0.timeseries.states:v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.states:int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.states:int_theta')

        theta_sol = sol.get_val('traj.phase0.timeseries.controls:theta')
        theta_sim = sim.get_val('traj.phase0.timeseries.controls:theta')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, tolerance=1.0E-2)

        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sol, theta_sol, tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sim, int_theta_sim, t_sim, theta_sim, tolerance=1.0E-2)

    def test_integrate_control_gl(self):
        self._test_integrate_control(dm.GaussLobatto)

    def test_integrate_control_radau(self):
        self._test_integrate_control(dm.Radau)

    def test_integrate_control_rate_gl(self):
        self._test_integrate_control_rate(dm.GaussLobatto)

    def test_integrate_control_rate_radau(self):
        self._test_integrate_control_rate(dm.Radau)

    def test_integrate_control_rate2_gl(self):
        self._test_integrate_control_rate2(dm.GaussLobatto)

    def test_integrate_control_rate2_radau(self):
        self._test_integrate_control_rate2(dm.Radau)


@use_tempdirs
class TestIntegratePolynomialControl(unittest.TestCase):

    def _test_integrate_polynomial_control(self, transcription):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=20, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta', fix_initial=False, rate_source='theta_rate',
                        targets=['theta'])

        # Define theta as a control.
        phase.add_polynomial_control(name='theta_rate', order=11, units='rad/s', shape=(1,), targets=None)

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        p.set_val('traj.phase0.t_initial', 0.0, units='s')
        p.set_val('traj.phase0.t_duration', 5.0, units='s')

        p.set_val('traj.phase0.states:x',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:y',
                  phase.interpolate(ys=[10, 5], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:v',
                  phase.interpolate(ys=[0, 5], nodes='state_input'),
                  units='m/s')

        p.set_val('traj.phase0.states:int_theta',
                  phase.interpolate(ys=[0.1, 45], nodes='state_input'),
                  units='deg')

        p.set_val('traj.phase0.polynomial_controls:theta_rate', 10.0, units='deg/s')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=True)

        sol = om.CaseReader('dymos_solution.db').get_case('final')
        sim = om.CaseReader('dymos_simulation.db').get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.states:x')
        x_sim = sim.get_val('traj.phase0.timeseries.states:x')

        y_sol = sol.get_val('traj.phase0.timeseries.states:y')
        y_sim = sim.get_val('traj.phase0.timeseries.states:y')

        v_sol = sol.get_val('traj.phase0.timeseries.states:v')
        v_sim = sim.get_val('traj.phase0.timeseries.states:v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.states:int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.states:int_theta')

        theta_rate_sol = sol.get_val('traj.phase0.timeseries.polynomial_controls:theta_rate')
        theta_rate_sim = sim.get_val('traj.phase0.timeseries.polynomial_controls:theta_rate')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, theta_rate_sol, t_sim, theta_rate_sim, tolerance=1.0E-2)

    def _test_integrate_polynomial_control_rate(self, transcription):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=5))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta', fix_initial=False, rate_source='theta_rate',
                        targets=['theta'])

        # Define theta as a control.
        phase.add_polynomial_control(name='theta', order=11, units='rad', shape=(1,), targets=None)

        # Force the initial value of the theta polynomial control to equal the initial value of the theta state.
        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='theta', var_b='int_theta',
                                    loc_a='initial', loc_b='initial')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        p.set_val('traj.phase0.t_initial', 0.0, units='s')
        p.set_val('traj.phase0.t_duration', 5.0, units='s')

        p.set_val('traj.phase0.states:x',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:y',
                  phase.interpolate(ys=[10, 5], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:v',
                  phase.interpolate(ys=[0, 5], nodes='state_input'),
                  units='m/s')

        p.set_val('traj.phase0.states:int_theta',
                  phase.interpolate(ys=[0.1, 45], nodes='state_input'),
                  units='deg')

        p.set_val('traj.phase0.polynomial_controls:theta', 45, units='deg')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=False)

        sol = om.CaseReader('dymos_solution.db').get_case('final')
        sim = om.CaseReader('dymos_simulation.db').get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.states:x')
        x_sim = sim.get_val('traj.phase0.timeseries.states:x')

        y_sol = sol.get_val('traj.phase0.timeseries.states:y')
        y_sim = sim.get_val('traj.phase0.timeseries.states:y')

        v_sol = sol.get_val('traj.phase0.timeseries.states:v')
        v_sim = sim.get_val('traj.phase0.timeseries.states:v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.states:int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.states:int_theta')

        theta_sol = sol.get_val('traj.phase0.timeseries.polynomial_controls:theta')
        theta_sim = sim.get_val('traj.phase0.timeseries.polynomial_controls:theta')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, tolerance=1.0E-3)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim,
                                     tolerance=1.0E-3)

        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sol, theta_sol, tolerance=1.0E-3)
        # assert_timeseries_near_equal(t_sol, int_theta_sim, t_sol, theta_sim, tolerance=1.0E-3)

    def _test_integrate_polynomial_control_rate2(self, transcription):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=20, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta_dot', fix_initial=False, rate_source='theta_rate2')
        phase.add_state('int_theta', fix_initial=False, rate_source='int_theta_dot',
                        targets=['theta'])

        # Define theta as a control.
        phase.add_polynomial_control(name='theta', order=11, units='rad', shape=(1,), targets=None)

        # Force the initial value of the theta polynomial control to equal the initial value of the theta state.
        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='theta', var_b='int_theta',
                                    loc_a='initial', loc_b='initial')

        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='int_theta_dot', var_b='theta_rate',
                                    loc_a='initial', loc_b='initial',
                                    units='rad/s')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.pyOptSparseDriver(optimizer='SLSQP')

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        p.set_val('traj.phase0.t_initial', 0.0, units='s')
        p.set_val('traj.phase0.t_duration', 5.0, units='s')

        p.set_val('traj.phase0.states:x',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:y',
                  phase.interpolate(ys=[10, 5], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:v',
                  phase.interpolate(ys=[0, 5], nodes='state_input'),
                  units='m/s')

        p.set_val('traj.phase0.states:int_theta',
                  phase.interpolate(ys=[0.1, 45], nodes='state_input'),
                  units='deg')

        p.set_val('traj.phase0.states:int_theta_dot',
                  phase.interpolate(ys=[0.0, 0.0], nodes='state_input'),
                  units='deg/s')

        p.set_val('traj.phase0.polynomial_controls:theta', 45, units='deg')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=True)

        sol = om.CaseReader('dymos_solution.db').get_case('final')
        sim = om.CaseReader('dymos_simulation.db').get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.states:x')
        x_sim = sim.get_val('traj.phase0.timeseries.states:x')

        y_sol = sol.get_val('traj.phase0.timeseries.states:y')
        y_sim = sim.get_val('traj.phase0.timeseries.states:y')

        v_sol = sol.get_val('traj.phase0.timeseries.states:v')
        v_sim = sim.get_val('traj.phase0.timeseries.states:v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.states:int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.states:int_theta')

        theta_sol = sol.get_val('traj.phase0.timeseries.polynomial_controls:theta')
        theta_sim = sim.get_val('traj.phase0.timeseries.polynomial_controls:theta')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, tolerance=1.0E-2)

    def test_integrate_polynomial_control_gl(self):
        self._test_integrate_polynomial_control(dm.GaussLobatto)

    def test_integrate_polynomial_control_radau(self):
        self._test_integrate_polynomial_control(dm.Radau)

    def test_integrate_polynomial_control_rate_gl(self):
        self._test_integrate_polynomial_control_rate(dm.GaussLobatto)

    def test_integrate_polynomial_control_rate_radau(self):
        self._test_integrate_polynomial_control_rate(dm.Radau)

    def test_integrate_polynomial_control_rate2_gl(self):
        self._test_integrate_polynomial_control_rate2(dm.GaussLobatto)

    def test_integrate_polynomial_control_rate2_radau(self):
        self._test_integrate_polynomial_control_rate2(dm.Radau)


@use_tempdirs
class TestIntegrateTimeParamAndState(unittest.TestCase):

    def _test_transcription(self, transcription=dm.GaussLobatto):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=3,
                                                     solve_segments='forward'))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, fix_duration=True, units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True)
        phase.add_state('y', fix_initial=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_one', fix_initial=True, rate_source='one')
        phase.add_state('int_time', fix_initial=True, rate_source='time')
        phase.add_state('int_time_phase', fix_initial=True, rate_source='time_phase')
        phase.add_state('int_int_one', fix_initial=True, rate_source='int_one')

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi)

        # With no targets we must explicitly assign units and shape to this parameter.
        # Its only purpose is to be integrated as the rate source for a state.
        phase.add_parameter(name='one', opt=False, units=None, shape=(1,))

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        p.set_val('traj.phase0.t_initial', 0.0, units='s')
        p.set_val('traj.phase0.t_duration', 5.0, units='s')

        p.set_val('traj.phase0.parameters:one', 1.0)

        p.set_val('traj.phase0.states:x',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:y',
                  phase.interpolate(ys=[10, 5], nodes='state_input'),
                  units='m')

        p.set_val('traj.phase0.states:v',
                  phase.interpolate(ys=[0, 5], nodes='state_input'),
                  units='m/s')

        p.set_val('traj.phase0.controls:theta',
                  phase.interpolate(ys=[0.1, 45], nodes='control_input'),
                  units='deg')

        # Additional states to test rate sources
        p.set_val('traj.phase0.states:int_one',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='s')

        p.set_val('traj.phase0.states:int_time',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='s**2')

        p.set_val('traj.phase0.states:int_time_phase',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='s**2')

        p.set_val('traj.phase0.states:int_int_one',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='s**2')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True)

        time_sol = p.get_val('traj.phase0.timeseries.time')
        time_phase_sol = p.get_val('traj.phase0.timeseries.time_phase')
        int_one_sol = p.get_val('traj.phase0.timeseries.states:int_one')
        int_time_sol = p.get_val('traj.phase0.timeseries.states:int_time')
        int_time_phase_sol = p.get_val('traj.phase0.timeseries.states:int_time_phase')
        int_int_one_sol = p.get_val('traj.phase0.timeseries.states:int_int_one')

        time_sim = p.get_val('traj.phase0.timeseries.time')
        time_phase_sim = p.get_val('traj.phase0.timeseries.time_phase')
        int_one_sim = p.get_val('traj.phase0.timeseries.states:int_one')
        int_time_sim = p.get_val('traj.phase0.timeseries.states:int_time')
        int_time_phase_sim = p.get_val('traj.phase0.timeseries.states:int_time_phase')
        int_int_one_sim = p.get_val('traj.phase0.timeseries.states:int_int_one')

        # Integral of one should match time and time_phase in this case.
        assert_near_equal(int_one_sol, time_sol, tolerance=1.0E-12)
        assert_near_equal(int_one_sol, time_phase_sol, tolerance=1.0E-12)

        assert_near_equal(int_one_sim, time_sim, tolerance=1.0E-12)
        assert_near_equal(int_one_sim, time_phase_sim, tolerance=1.0E-12)

        # Integral of time and time_phase should be t**2/2
        assert_near_equal(time_sol, time_phase_sol, tolerance=1.0E-12)
        assert_near_equal(int_time_sol, time_sol**2/2, tolerance=1.0E-12)
        assert_near_equal(int_time_phase_sol, time_phase_sol**2/2, tolerance=1.0E-12)

        assert_near_equal(time_sim, time_phase_sim, tolerance=1.0E-12)
        assert_near_equal(int_time_sim, time_sim**2/2, tolerance=1.0E-12)
        assert_near_equal(int_time_phase_sim, time_phase_sim**2/2, tolerance=1.0E-12)

        # Double integral of one should be the same as the integral of time
        assert_near_equal(int_int_one_sol, int_time_sol, tolerance=1.0E-12)
        assert_near_equal(int_int_one_sim, int_time_sim, tolerance=1.0E-12)

        assert_timeseries_near_equal(time_sol, int_int_one_sol, time_sim, int_int_one_sim)

    def test_gl(self):
        self._test_transcription(transcription=dm.GaussLobatto)

    def test_radau(self):
        self._test_transcription(transcription=dm.Radau)



import os.path
from pprint import pprint
import shutil
import unittest

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
plt.style.use('ggplot')

from dymos.utils.doc_utils import save_for_docs
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.coloring import Coloring


def _view_coloring(coloring_file, show_sparsity_text=False, show_sparsity=True,
                   subjac_sparsity=False, color_var=None, show_meta=False):
    coloring = Coloring.load(coloring_file)
    if show_sparsity_text:
        coloring.display_txt()

    if show_sparsity:
        coloring.display()
        fig = plt.gcf()
        fig.set_size_inches(5.5, 5.5)
        fig.tight_layout()

    if subjac_sparsity:
        print("\nSubjacobian sparsity:")
        for tup in coloring._subjac_sparsity_iter():
            print("(%s, %s)\n   rows=%s\n   cols=%s" % tup[:4])
        print()

    if color_var is not None:
        fwd, rev = coloring.get_row_var_coloring(color_var)
        print("\nVar: %s  (fwd solves: %d,  rev solves: %d)\n" % (color_var, fwd, rev))

    if show_meta:
        print("\nColoring metadata:")
        pprint(coloring._meta)

    coloring.summary()


@use_tempdirs
class TestMinTimeClimbForDocs(unittest.TestCase):

    @save_for_docs
    def test_min_time_climb_for_docs_partial_coloring(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal

        import dymos as dm
        from dymos.examples.min_time_climb.doc.min_time_climb_ode_partial_coloring import MinTimeClimbODE

        for fd in (False, True):
            if fd:
                pc_options = (False, True)
            else:
                pc_options = (False,)
            for pc in pc_options:
                with self.subTest(f'Finite Differencing: {fd}  Partial Coloring: {pc}'):
                    print(f'Finite Differencing: {fd}  Partial Coloring: {pc}')

                    #
                    # Instantiate the problem and configure the optimization driver
                    #
                    p = om.Problem(model=om.Group())

                    p.driver = om.pyOptSparseDriver()
                    p.driver.options['optimizer'] = 'IPOPT'
                    p.driver.declare_coloring(tol=1.0E-12)
                    p.driver.opt_settings['max_iter'] = 500
                    p.driver.opt_settings['alpha_for_y'] = 'safer-min-dual-infeas'

                    #
                    # Instantiate the trajectory and phase
                    #
                    traj = dm.Trajectory()

                    phase = dm.Phase(ode_class=MinTimeClimbODE,
                                     ode_init_kwargs={'fd': fd, 'partial_coloring': pc},
                                     transcription=dm.GaussLobatto(num_segments=30))

                    traj.add_phase('phase0', phase)

                    p.model.add_subsystem('traj', traj)

                    #
                    # Set the options on the optimization variables
                    #
                    phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                                           duration_ref=100.0)

                    phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6,
                                    ref=1.0E3, defect_ref=1.0E3, units='m',
                                    rate_source='flight_dynamics.r_dot')

                    phase.add_state('h', fix_initial=True, lower=0, upper=20000.0,
                                    ref=1.0E2, defect_ref=1.0E2, units='m',
                                    rate_source='flight_dynamics.h_dot')

                    phase.add_state('v', fix_initial=True, lower=10.0,
                                    ref=1.0E2, defect_ref=1.0E2, units='m/s',
                                    rate_source='flight_dynamics.v_dot')

                    phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5,
                                    ref=1.0, defect_ref=1.0, units='rad',
                                    rate_source='flight_dynamics.gam_dot')

                    phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5,
                                    ref=1.0E3, defect_ref=1.0E3, units='kg',
                                    rate_source='prop.m_dot')

                    phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                                      rate_continuity=True, rate_continuity_scaler=100.0,
                                      rate2_continuity=False, targets=['alpha'])

                    phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
                    phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
                    phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

                    #
                    # Setup the boundary and path constraints
                    #
                    phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3)
                    phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0)
                    phase.add_boundary_constraint('gam', loc='final', equals=0.0)

                    phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
                    phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8)

                    # Minimize time at the end of the phase
                    phase.add_objective('time', loc='final', ref=100.0)

                    p.model.linear_solver = om.DirectSolver()

                    #
                    # Setup the problem and set the initial guess
                    #
                    p.setup()

                    p['traj.phase0.t_initial'] = 0.0
                    p['traj.phase0.t_duration'] = 500

                    p['traj.phase0.states:r'] = phase.interpolate(ys=[0.0, 50000.0], nodes='state_input')
                    p['traj.phase0.states:h'] = phase.interpolate(ys=[100.0, 20000.0], nodes='state_input')
                    p['traj.phase0.states:v'] = phase.interpolate(ys=[135.964, 283.159], nodes='state_input')
                    p['traj.phase0.states:gam'] = phase.interpolate(ys=[0.0, 0.0], nodes='state_input')
                    p['traj.phase0.states:m'] = phase.interpolate(ys=[19030.468, 10000.], nodes='state_input')
                    p['traj.phase0.controls:alpha'] = phase.interpolate(ys=[0.0, 0.0], nodes='control_input')

                    #
                    # Solve for the optimal trajectory
                    #
                    dm.run_problem(p)

                    #
                    # This code is intended to save the coloring plots for the documentation.
                    # In practice, use the command line interface to view these files instead:
                    # `openmdao view_coloring coloring_files/total_coloring.pkl --view`
                    #
                    stfd = '_fd' if fd else ''
                    stpc = '_pc' if pc else ''
                    coloring_dir = f'coloring_files{stfd}{stpc}'
                    if fd or pc:
                        if os.path.exists(coloring_dir):
                            shutil.rmtree(coloring_dir)
                        shutil.move('coloring_files', coloring_dir)

                    _view_coloring(os.path.join(coloring_dir, 'total_coloring.pkl'))

                    #
                    # Test the results
                    #
                    assert_near_equal(p.get_val('traj.phase0.t_duration'), 321.0, tolerance=1.0E-1)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.transcriptions.runge_kutta.components import RungeKuttaStateContinuityIterGroup
from dymos.transcriptions.runge_kutta.test.rk_test_ode import TestODE


@use_tempdirs
class TestRungeKuttaContinuityIterGroup(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_continuity_comp_no_iteration(self):
        num_seg = 4
        state_options = {'y': {'shape': (1,), 'units': 'm', 'targets': ['y'], 'fix_initial': True,
                               'fix_final': False, 'propagation': 'forward', 'defect_scaler': None,
                               'defect_ref': None, 'lower': None, 'upper': None,
                               'connected_initial': False}}

        class RKGroup(om.Group):

            def setup(self):

                ivc = self.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

                ivc.add_output('time', val=np.array([0.00, 0.25, 0.25, 0.50,
                                                     0.50, 0.75, 0.75, 1.00,
                                                     1.00, 1.25, 1.25, 1.50,
                                                     1.50, 1.75, 1.75, 2.00]), units='s')

                ivc.add_output('h', val=np.array([0.5, 0.5, 0.5, 0.5]), units='s')

                self.add_subsystem('cnty_iter_group',
                                   RungeKuttaStateContinuityIterGroup(
                                       num_segments=num_seg,
                                       method='RK4',
                                       state_options=state_options,
                                       time_units='s',
                                       ode_class=TestODE,
                                       ode_init_kwargs={},
                                       k_solver_class=om.NonlinearRunOnce),
                                   promotes_outputs=['states:*'])

                self.connect('h', 'cnty_iter_group.h')
                self.connect('time', 'cnty_iter_group.ode.t')

                src_idxs = np.arange(16, dtype=int).reshape((num_seg, 4, 1))
                self.connect('cnty_iter_group.ode.ydot', 'cnty_iter_group.k_comp.f:y',
                             src_indices=src_idxs, flat_src_indices=True)

                self.nonlinear_solver = om.NonlinearRunOnce()
                self.linear_solver = om.DirectSolver()

            def configure(self):
                self.cnty_iter_group.configure_io()

        p = om.Problem(model=RKGroup())

        p.setup(check=True, force_alloc_complex=True)

        p['states:y'] = np.array([[0.50000000],
                                  [1.425130208333333],
                                  [2.639602661132812],
                                  [4.006818970044454],
                                  [5.301605229265987]])

        p['cnty_iter_group.k_comp.k:y'] = np.array([[[0.75000000],
                                                     [0.90625000],
                                                     [0.94531250],
                                                     [1.09765625]],

                                                    [[1.087565104166667],
                                                     [1.203206380208333],
                                                     [1.232116699218750],
                                                     [1.328623453776042]],

                                                    [[1.319801330566406],
                                                     [1.368501663208008],
                                                     [1.380676746368408],
                                                     [1.385139703750610]],

                                                    [[1.378409485022227],
                                                     [1.316761856277783],
                                                     [1.301349949091673],
                                                     [1.154084459568063]]])

        p.run_model()
        p.model.run_apply_nonlinear()

        # Test that the residuals of the states are the expected values
        outputs = p.model.list_outputs(print_arrays=True, residuals=True, out_stream=None)

        expected_resids = np.zeros((num_seg + 1, 1))

        op_dict = dict([op for op in outputs])
        assert_near_equal(op_dict['cnty_iter_group.continuity_comp.states:y']['resids'],
                          expected_resids)

        # Test the partials
        cpd = p.check_partials(method='cs', out_stream=None)

        J_fwd = cpd['cnty_iter_group.continuity_comp']['states:y', 'state_integrals:y']['J_fwd']
        J_fd = cpd['cnty_iter_group.continuity_comp']['states:y', 'state_integrals:y']['J_fd']
        assert_near_equal(J_fwd, J_fd)

        J_fwd = cpd['cnty_iter_group.continuity_comp']['states:y', 'states:y']['J_fwd']
        J_fd = cpd['cnty_iter_group.continuity_comp']['states:y', 'states:y']['J_fd']

        J_fd[0, 0] = -1.0

        assert_near_equal(J_fwd, J_fd)

    def test_continuity_comp_newtonsolver(self):
        num_seg = 4
        state_options = {'y': {'shape': (1,), 'units': 'm', 'targets': ['y'], 'fix_initial': True,
                               'fix_final': False, 'propagation': 'forward', 'defect_scaler': None,
                               'defect_ref': 1.0, 'lower': None, 'upper': None,
                               'connected_initial': False}}

        class RKGroup(om.Group):

            def setup(self):
                ivc = self.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

                ivc.add_output('time', val=np.array([0.00, 0.25, 0.25, 0.50,
                                                     0.50, 0.75, 0.75, 1.00,
                                                     1.00, 1.25, 1.25, 1.50,
                                                     1.50, 1.75, 1.75, 2.00]), units='s')

                ivc.add_output('h', val=np.array([0.5, 0.5, 0.5, 0.5]), units='s')

                self.add_subsystem('cnty_iter_group',
                                   RungeKuttaStateContinuityIterGroup(
                                       num_segments=num_seg,
                                       method='RK4',
                                       state_options=state_options,
                                       time_units='s',
                                       ode_class=TestODE,
                                       ode_init_kwargs={},
                                       k_solver_class=None),
                                   promotes_outputs=['states:*'])

                self.connect('h', 'cnty_iter_group.h')
                self.connect('time', 'cnty_iter_group.ode.t')

                src_idxs = np.arange(16, dtype=int).reshape((num_seg, 4, 1))
                self.connect('cnty_iter_group.ode.ydot', 'cnty_iter_group.k_comp.f:y',
                             src_indices=src_idxs, flat_src_indices=True)

                self.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
                self.linear_solver = om.DirectSolver()

            def configure(self):
                self.cnty_iter_group.configure_io()

        p = om.Problem(model=RKGroup())

        p.setup(check=True, force_alloc_complex=True)

        # Set the initial value (and all other values) of y
        p['states:y'][...] = 0.50

        p.run_model()
        # p.model.run_apply_nonlinear()

        # Test that the residuals of the states are the expected values
        outputs = p.model.list_outputs(print_arrays=True, residuals=True, out_stream=None)

        expected_resids = np.zeros((num_seg + 1, 1))

        op_dict = dict([op for op in outputs])
        assert_near_equal(op_dict['cnty_iter_group.continuity_comp.states:y']['resids'],
                          expected_resids)

        # Test the partials
        cpd = p.check_partials(method='cs', out_stream=None)

        J_fwd = cpd['cnty_iter_group.continuity_comp']['states:y', 'state_integrals:y']['J_fwd']
        J_fd = cpd['cnty_iter_group.continuity_comp']['states:y', 'state_integrals:y']['J_fd']
        assert_near_equal(J_fwd, J_fd)

        J_fwd = cpd['cnty_iter_group.continuity_comp']['states:y', 'states:y']['J_fwd']
        J_fd = cpd['cnty_iter_group.continuity_comp']['states:y', 'states:y']['J_fd']

        J_fd[0, 0] = -1.0

        assert_near_equal(J_fwd, J_fd)

if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import openmdao.api as om
import dymos as dm


# @dm.declare_time(targets=['t'], units='s')
# @dm.declare_state('y', targets=['y'], rate_source='ydot', units='m')
class TestODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        self.add_input('t', val=np.ones(self.options['num_nodes']), units='s')
        self.add_input('y', val=np.ones(self.options['num_nodes']), units='m')
        self.add_output('ydot', val=np.ones(self.options['num_nodes']), units='m/s')

        ar = np.arange(self.options['num_nodes'])
        self.declare_partials(of='ydot', wrt='t', rows=ar, cols=ar)
        self.declare_partials(of='ydot', wrt='y', rows=ar, cols=ar, val=1.0)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        t = inputs['t']
        y = inputs['y']
        outputs['ydot'] = y - t ** 2 + 1

    def compute_partials(self, inputs, partials):
        partials['ydot', 't'] = -2 * inputs['t']


def _test_ode_solution(y, t):
    return t**2 + 2*t + 1 - 0.5 * np.exp(t)



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.transcriptions.runge_kutta.components.runge_kutta_state_advance_comp import \
    RungeKuttaStateAdvanceComp
from dymos.utils.testing_utils import assert_check_partials

# Modify class so we can run it standalone.
from dymos.utils.misc import CompWrapperConfig
RungeKuttaStateAdvanceComp = CompWrapperConfig(RungeKuttaStateAdvanceComp)


@use_tempdirs
class TestRKStateAdvanceComp(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_rk_state_advance_comp_rk4_scalar(self):
        state_options = {'y': {'shape': (1,), 'units': 'm'}}

        p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

        ivc.add_output('k:y', shape=(4, 4, 1), units='m')
        ivc.add_output('y0', shape=(4, 1), units='m')

        p.model.add_subsystem('c',
                              RungeKuttaStateAdvanceComp(num_segments=4, method='RK4',
                                                         state_options=state_options))

        p.model.connect('k:y', 'c.k:y')
        p.model.connect('y0', 'c.initial_states_per_seg:y')

        p.setup(check=True, force_alloc_complex=True)

        p['y0'] = np.array([[0.5, 1.425130208333333, 2.639602661132812, 4.006818970044454]]).T
        p['k:y'] = np.array([[[0.75000000],
                              [0.90625000],
                              [0.94531250],
                              [1.09765625]],

                             [[1.08756510],
                              [1.20320638],
                              [1.23211670],
                              [1.32862345]],

                             [[1.31980133],
                              [1.36850166],
                              [1.38067675],
                              [1.38513970]],

                             [[1.37840949],
                              [1.31676186],
                              [1.30134995],
                              [1.15408446]]])

        p.run_model()

        assert_near_equal(p.get_val('c.final_states:y'),
                          np.array([[1.425130208333333],
                                    [2.639602661132812],
                                    [4.006818970044454],
                                    [5.301605229265987]]),
                          tolerance=1.0E-9)

        np.set_printoptions(linewidth=1024)
        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)

    def test_rk_state_advance_comp_rk4_vector(self):
        state_options = {'y': {'shape': (2,), 'units': 'm'}}

        p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

        ivc.add_output('k:y', shape=(2, 4, 2), units='m')
        ivc.add_output('y0', shape=(2, 2), units='m')

        p.model.add_subsystem('c',
                              RungeKuttaStateAdvanceComp(num_segments=2, method='RK4',
                                                         state_options=state_options))

        p.model.connect('k:y', 'c.k:y')
        p.model.connect('y0', 'c.initial_states_per_seg:y')

        p.setup(check=True, force_alloc_complex=True)

        p['y0'] = [[0.5, 1.425130208333333], [2.639602661132812, 4.006818970044454]]
        p['k:y'][0, :, 0] = [0.75, 0.90625, 0.9453125, 1.09765625]
        p['k:y'][0, :, 1] = [1.08756510, 1.20320638, 1.23211670, 1.32862345]
        p['k:y'][1, :, 0] = [1.31980133, 1.36850166, 1.38067675, 1.38513970]
        p['k:y'][1, :, 1] = [1.37840949, 1.31676186, 1.30134995, 1.15408446]

        p.run_model()

        assert_near_equal(actual=p.get_val('c.final_states:y'),
                          desired=[[1.425130208333333, 2.639602661132812],
                                   [4.006818970044454, 5.301605229265987]],
                          tolerance=1.0E-9)

        np.set_printoptions(linewidth=1024)
        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)

    def test_rk_state_advance_comp_rk4_matrix(self):
        state_options = {'y': {'shape': (2, 2), 'units': 'm'}}

        p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

        ivc.add_output('k:y', shape=(4, 2, 2), units='m')
        ivc.add_output('y0', shape=(2, 2), units='m')

        p.model.add_subsystem('c',
                              RungeKuttaStateAdvanceComp(num_segments=1,
                                                         method='RK4',
                                                         state_options=state_options))

        p.model.connect('k:y', 'c.k:y')
        p.model.connect('y0', 'c.initial_states_per_seg:y')

        p.setup(check=True, force_alloc_complex=True)

        p['y0'] = [[0.5, 1.425130208333333], [2.639602661132812, 4.006818970044454]]
        p['k:y'] = np.array([
                            [[0.75, 1.087565104166667],
                             [1.319801330566406, 1.378409485022227]],

                            [[0.90625, 1.203206380208333],
                             [1.368501663208008,  1.316761856277783]],

                            [[0.9453125, 1.23211669921875],
                             [1.380676746368408,  1.301349949091673]],

                            [[1.09765625, 1.328623453776042],
                             [1.385139703750610,  1.154084459568063]]])

        p.run_model()

        assert_near_equal(p.get_val('c.final_states:y'),
                          [[[1.425130208333333, 2.639602661132812],
                           [4.006818970044454, 5.301605229265987]]])

        np.set_printoptions(linewidth=1024)
        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.transcriptions.runge_kutta.components.runge_kutta_stepsize_comp import RungeKuttaStepsizeComp
from dymos.utils.testing_utils import assert_check_partials

from dymos.utils.misc import CompWrapperConfig
RungeKuttaStepsizeComp = CompWrapperConfig(RungeKuttaStepsizeComp)


@use_tempdirs
class TestRKStepsizeComp(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_rk_stepsize_comp(self):
        p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

        ivc.add_output('t_duration', shape=(1,), units='s')

        p.model.add_subsystem('c', RungeKuttaStepsizeComp(num_segments=4,
                                                          seg_rel_lengths=[1, 1, 1, 1],
                                                          time_units='s'))

        p.model.connect('t_duration', 'c.t_duration')

        p.setup(check=True, force_alloc_complex=True)

        p['t_duration'] = 2.0

        np.set_printoptions(linewidth=1024)

        p.run_model()

        expected = np.array([0.5, 0.5, 0.5, 0.5])

        assert_near_equal(p.get_val('c.h'), expected, tolerance=1.0E-9)

        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)

    def test_rk_stepsize_comp_nonuniform(self):
        p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

        ivc.add_output('t_duration', shape=(1,), units='s')

        p.model.add_subsystem('c', RungeKuttaStepsizeComp(num_segments=4,
                                                          seg_rel_lengths=[1, 1, 0.5, 0.25],
                                                          time_units='s'))

        p.model.connect('t_duration', 'c.t_duration')

        p.setup(check=True, force_alloc_complex=True)

        p['t_duration'] = 5.5

        np.set_printoptions(linewidth=1024)

        p.run_model()

        expected = np.array([2.0, 2.0, 1.0, 0.5])

        assert_near_equal(p.get_val('c.h'), expected, tolerance=1.0E-9)

        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)



from fnmatch import filter
import warnings

import numpy as np

import openmdao.api as om

from ..transcription_base import TranscriptionBase
from .components import SegmentSimulationComp, SegmentStateMuxComp, \
    SolveIVPControlGroup, SolveIVPPolynomialControlGroup, SolveIVPTimeseriesOutputComp
from ..common import TimeComp
from ...utils.misc import get_rate_units, get_target_metadata, get_source_metadata, \
    _unspecified
from ...utils.introspection import get_targets
from ...utils.indexing import get_src_indices_by_row


class SolveIVP(TranscriptionBase):
    """
    The SolveIVP Transcription class.

    SolveIVP transcription in Dymos uses the scipy.simulate.solve_ivp method to explicitly integrate
    the states from the phase initial time to the phase final time.

    SolveIVP transcription does not currently support optimization since it does not propagate
    analytic derivatives through the ODE.

    Parameters
    ----------
    grid_data : GridData
        Grid data for this phases.
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, grid_data=None, **kwargs):
        super(SolveIVP, self).__init__(**kwargs)
        self.grid_data = grid_data
        self._rhs_source = 'ode'

    def initialize(self):
        """
        Declare transcription options.
        """
        super(SolveIVP, self).initialize()

        self.options.declare('method', default='RK45', values=('RK45', 'RK23', 'BDF'),
                             desc='The integrator used within scipy.integrate.solve_ivp. Currently '
                                  'supports \'RK45\', \'RK23\', and \'BDF\'.')

        self.options.declare('atol', default=1.0E-6, types=(float,),
                             desc='Absolute tolerance passed to scipy.integrate.solve_ivp.')

        self.options.declare('rtol', default=1.0E-6, types=(float,),
                             desc='Relative tolerance passed to scipy.integrate.solve_ivp.')

        self.options.declare('output_nodes_per_seg', default=None, types=(int,), allow_none=True,
                             desc='If None, results are provided at the all nodes within each'
                                  'segment.  If an int (n) then results are provided at n '
                                  'equally distributed points in time within each segment.')

    def init_grid(self):
        """
        Setup the GridData object for the Transcription.
        """
        pass

    def setup_time(self, phase):
        """
        Setup the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        time_options = phase.time_options
        time_units = time_options['units']
        num_seg = self.grid_data.num_segments
        grid_data = self.grid_data
        output_nodes_per_seg = self.options['output_nodes_per_seg']

        time_options['input_initial'] = False  # True can break simulation
        time_options['input_duration'] = False

        super(SolveIVP, self).setup_time(phase)

        if output_nodes_per_seg is None:
            # Case 1:  Compute times at 'all' node set.
            num_nodes = grid_data.num_nodes
            node_ptau = grid_data.node_ptau
            node_dptau_dstau = grid_data.node_dptau_dstau
        else:
            # Case 2:  Compute times at n equally distributed points per segment.
            num_nodes = num_seg * output_nodes_per_seg
            node_stau = np.linspace(-1, 1, output_nodes_per_seg)
            node_ptau = np.empty(0, )
            node_dptau_dstau = np.empty(0, )
            # Append our nodes in phase tau space
            for iseg in range(num_seg):
                v0 = grid_data.segment_ends[iseg]
                v1 = grid_data.segment_ends[iseg + 1]
                node_ptau = np.concatenate((node_ptau, v0 + 0.5 * (node_stau + 1) * (v1 - v0)))
                node_dptau_dstau = np.concatenate((node_dptau_dstau,
                                                   0.5 * (v1 - v0) * np.ones_like(node_stau)))

        time_comp = TimeComp(num_nodes=num_nodes, node_ptau=node_ptau,
                             node_dptau_dstau=node_dptau_dstau, units=time_units)

        phase.add_subsystem('time', time_comp, promotes=['*'])

    def configure_time(self, phase):
        """
        Configure the inputs/outputs on the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(SolveIVP, self).configure_time(phase)
        num_seg = self.grid_data.num_segments
        grid_data = self.grid_data
        output_nodes_per_seg = self.options['output_nodes_per_seg']

        phase.time.configure_io()

        for i in range(num_seg):
            phase.connect('t_initial', f'segment_{i}.t_initial')
            phase.connect('t_duration', f'segment_{i}.t_duration')
            if output_nodes_per_seg is None:
                i1, i2 = grid_data.subset_segment_indices['all'][i, :]
                src_idxs = grid_data.subset_node_indices['all'][i1:i2]
            else:
                src_idxs = np.arange(i * output_nodes_per_seg, output_nodes_per_seg * (i + 1),
                                     dtype=int)
            phase.connect('time', f'segment_{i}.time', src_indices=src_idxs)
            phase.connect('time_phase', f'segment_{i}.time_phase', src_indices=src_idxs)

        options = phase.time_options

        # The tuples here are (name, user_specified_targets, dynamic)
        for name, usr_tgts, dynamic in [('time', options['targets'], True),
                                        ('time_phase', options['time_phase_targets'], True),
                                        ('t_initial', options['t_initial_targets'], False),
                                        ('t_duration', options['t_duration_targets'], False)]:

            targets = get_targets(phase.ode, name=name, user_targets=usr_tgts)
            if targets:
                phase.connect(name, [f'ode.{t}' for t in targets])

    def setup_states(self, phase):
        """
        Setup the states for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        phase.add_subsystem('indep_states', om.IndepVarComp(),
                            promotes_outputs=['*'])

    def configure_states(self, phase):
        """
        Configure state connections post-introspection.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        num_seg = self.grid_data.num_segments

        for state_name, options in phase.state_options.items():
            phase.indep_states.add_output(f'initial_states:{state_name}',
                                          val=np.ones(((1,) + options['shape'])),
                                          units=options['units'])

        for state_name, options in phase.state_options.items():
            # Connect the initial state to the first segment
            src_idxs = get_src_indices_by_row([0], options['shape'])

            phase.connect(f'initial_states:{state_name}',
                          f'segment_0.initial_states:{state_name}',
                          src_indices=src_idxs, flat_src_indices=True)

            phase.connect(f'segment_0.states:{state_name}',
                          f'state_mux_comp.segment_0_states:{state_name}')

            targets = get_targets(ode=phase.ode, name=state_name, user_targets=options['targets'])

            if targets:
                phase.connect(f'state_mux_comp.states:{state_name}',
                              [f'ode.{t}' for t in targets])

            # Connect the final state in segment n to the initial state in segment n + 1
            for i in range(1, num_seg):
                if self.options['output_nodes_per_seg'] is None:
                    nnps_i = self.grid_data.subset_num_nodes_per_segment['all'][i]
                else:
                    nnps_i = self.options['output_nodes_per_seg']

                src_idxs = get_src_indices_by_row([nnps_i-1], shape=options['shape'])
                phase.connect(f'segment_{i - 1}.states:{state_name}',
                              f'segment_{i}.initial_states:{state_name}',
                              src_indices=src_idxs, flat_src_indices=True)

                phase.connect(f'segment_{i}.states:{state_name}',
                              f'state_mux_comp.segment_{i}_states:{state_name}')

    def setup_ode(self, phase):
        """
        Setup the ode for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data
        num_seg = gd.num_segments

        segments_group = phase.add_subsystem(name='segments', subsys=om.Group(),
                                             promotes_outputs=['*'], promotes_inputs=['*'])

        for i in range(num_seg):
            seg_i_comp = SegmentSimulationComp(
                index=i,
                method=self.options['method'],
                atol=self.options['atol'],
                rtol=self.options['rtol'],
                grid_data=self.grid_data,
                ode_class=phase.options['ode_class'],
                ode_init_kwargs=phase.options['ode_init_kwargs'],
                time_options=phase.time_options,
                state_options=phase.state_options,
                control_options=phase.control_options,
                polynomial_control_options=phase.polynomial_control_options,
                parameter_options=phase.parameter_options,
                output_nodes_per_seg=self.options['output_nodes_per_seg'])

            segments_group.add_subsystem(f'segment_{i}', subsys=seg_i_comp)

        # scipy.integrate.solve_ivp does not actually evaluate the ODE at the desired output points,
        # but just returns the time and interpolated integrated state values there instead. We need
        # to instantiate a second ODE group that will call the ODE at those points so that we can
        # accurately obtain timeseries for ODE outputs.
        phase.add_subsystem('state_mux_comp',
                            SegmentStateMuxComp(grid_data=gd, state_options=phase.state_options,
                                                output_nodes_per_seg=self.options['output_nodes_per_seg']))

        if self.options['output_nodes_per_seg'] is None:
            self.num_output_nodes = gd.subset_num_nodes['all']
        else:
            self.num_output_nodes = num_seg * self.options['output_nodes_per_seg']

        phase.add_subsystem('ode', phase.options['ode_class'](num_nodes=self.num_output_nodes,
                                                              **phase.options['ode_init_kwargs']))

    def configure_ode(self, phase):
        """
        Create connections to the introspected states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data
        num_seg = gd.num_segments

        for i in range(num_seg):
            seg_comp = phase.segments._get_subsystem(f'segment_{i}')
            seg_comp.configure_io()

    def setup_controls(self, phase):
        """
        Setup the control group.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        output_nodes_per_seg = self.options['output_nodes_per_seg']

        phase._check_control_options()

        if phase.control_options:
            control_group = SolveIVPControlGroup(control_options=phase.control_options,
                                                 time_units=phase.time_options['units'],
                                                 grid_data=self.grid_data,
                                                 output_nodes_per_seg=output_nodes_per_seg)

            phase.add_subsystem('control_group',
                                subsys=control_group,
                                promotes=['controls:*', 'control_values:*', 'control_values_all:*',
                                          'control_rates:*'])

    def configure_controls(self, phase):
        """
        Configure the inputs/outputs for the controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        ode = phase._get_subsystem(self._rhs_source)

        # Interrogate shapes and units.
        for name, options in phase.control_options.items():

            shape, units = get_target_metadata(ode, name=name,
                                               user_targets=options['targets'],
                                               user_units=options['units'],
                                               user_shape=options['shape'],
                                               control_rate=True)

            options['units'] = units
            options['shape'] = shape

        grid_data = self.grid_data

        if phase.control_options:
            phase.control_group.configure_io()
            phase.connect('dt_dstau', 'control_group.dt_dstau')

        for name, options in phase.control_options.items():
            for i in range(grid_data.num_segments):
                i1, i2 = grid_data.subset_segment_indices['control_disc'][i, :]
                seg_idxs = grid_data.subset_node_indices['control_disc'][i1:i2]
                src_idxs = get_src_indices_by_row(row_idxs=seg_idxs, shape=options['shape'])
                phase.connect(src_name=f'control_values_all:{name}',
                              tgt_name=f'segment_{i}.controls:{name}',
                              src_indices=src_idxs, flat_src_indices=True)

            targets = get_targets(ode=phase.ode, name=name, user_targets=options['targets'])
            if targets:
                phase.connect(f'control_values:{name}', [f'ode.{t}' for t in targets])

            targets = get_targets(ode=phase.ode, name=f'{name}_rate',
                                  user_targets=options['rate_targets'])
            if targets:
                phase.connect(f'control_rates:{name}_rate',
                              [f'ode.{t}' for t in targets])

            targets = get_targets(ode=phase.ode, name=f'{name}_rate2',
                                  user_targets=options['rate2_targets'])
            if targets:
                phase.connect(f'control_rates:{name}_rate2',
                              [f'ode.{t}' for t in targets])

    def setup_polynomial_controls(self, phase):
        """
        Adds the polynomial control group to the model if any polynomial controls are present.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        if phase.polynomial_control_options:
            sys = SolveIVPPolynomialControlGroup(grid_data=self.grid_data,
                                                 polynomial_control_options=phase.polynomial_control_options,
                                                 time_units=phase.time_options['units'],
                                                 output_nodes_per_seg=self.options['output_nodes_per_seg'])
            phase.add_subsystem('polynomial_control_group', subsys=sys,
                                promotes_inputs=['*'], promotes_outputs=['*'])

    def configure_polynomial_controls(self, phase):
        """
        Configure the inputs/outputs for the polynomial controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        # In transcription_base, we get the control units/shape from the target, and then call
        # configure on the control_group.
        super(SolveIVP, self).configure_polynomial_controls(phase)

        # Additional connections.
        for name, options in phase.polynomial_control_options.items():

            for iseg in range(self.grid_data.num_segments):
                phase.connect(src_name=f'polynomial_controls:{name}',
                              tgt_name=f'segment_{iseg}.polynomial_controls:{name}')

            targets = get_targets(ode=phase.ode, name=name, user_targets=options['targets'])
            if targets:
                phase.connect(f'polynomial_control_values:{name}', [f'ode.{t}' for t in targets])

            targets = get_targets(ode=phase.ode, name=f'{name}_rate',
                                  user_targets=options['rate_targets'])
            if targets:
                phase.connect(f'polynomial_control_rates:{name}_rate',
                              [f'ode.{t}' for t in targets])

            targets = get_targets(ode=phase.ode, name=f'{name}_rate2',
                                  user_targets=options['rate2_targets'])
            if targets:
                phase.connect(f'polynomial_control_rates:{name}_rate2',
                              [f'ode.{t}' for t in targets])

    def configure_parameters(self, phase):
        """
        Configure parameter promotion.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(SolveIVP, self).configure_parameters(phase)

        gd = self.grid_data

        # We also need to take care of the segments.
        segs = phase._get_subsystem('segments')

        for name, options in phase.parameter_options.items():
            prom_name = f'parameters:{name}'
            shape, units = get_target_metadata(phase.ode, name=name,
                                               user_targets=options['targets'],
                                               user_shape=options['shape'],
                                               user_units=options['units'])
            options['units'] = units
            options['shape'] = shape

            for i in range(gd.num_segments):
                seg_comp = segs._get_subsystem(f'segment_{i}')
                seg_comp.add_input(name=prom_name, val=np.ones(shape), units=units,
                                   desc=f'values of parameter {name}.')
                segs.promotes(f'segment_{i}', inputs=[prom_name])

    def setup_defects(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_defects(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_objective(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def setup_path_constraints(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_path_constraints(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def setup_boundary_constraints(self, loc, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        loc : str
            The kind of boundary constraints being setup.  Must be one of 'initial' or 'final'.
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_boundary_constraints(self, loc, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        loc : str
            The kind of boundary constraints being setup.  Must be one of 'initial' or 'final'.
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def setup_solvers(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_solvers(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def setup_timeseries_outputs(self, phase):
        """
        Setup the timeseries for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data

        timeseries_comp = \
            SolveIVPTimeseriesOutputComp(input_grid_data=gd,
                                         output_nodes_per_seg=self.options['output_nodes_per_seg'])

        phase.add_subsystem('timeseries', subsys=timeseries_comp)

    def configure_timeseries_outputs(self, phase):
        """
        Create connections from time series to all post-introspection sources.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data
        num_seg = gd.num_segments
        output_nodes_per_seg = self.options['output_nodes_per_seg']
        time_units = phase.time_options['units']

        timeseries_name = 'timeseries'
        timeseries_comp = phase._get_subsystem(timeseries_name)

        timeseries_comp._add_output_configure('time', shape=(1,), units=time_units, desc='time')
        timeseries_comp._add_output_configure('time_phase', shape=(1,), units=time_units,
                                              desc='elapsed phase time')

        phase.connect(src_name='time', tgt_name='timeseries.all_values:time')

        phase.connect(src_name='time_phase', tgt_name='timeseries.all_values:time_phase')

        for name, options in phase.state_options.items():

            timeseries_comp._add_output_configure(f'states:{name}',
                                                  shape=options['shape'],
                                                  units=options['units'],
                                                  desc=options['desc'])

            timeseries_comp._add_output_configure(f'state_rates:{name}',
                                                  shape=options['shape'],
                                                  units=get_rate_units(options['units'],
                                                                       time_units, deriv=1),
                                                  desc=f'first time-derivative of state {name}')

            phase.connect(src_name=f'state_mux_comp.states:{name}',
                          tgt_name=f'timeseries.all_values:states:{name}')

            rate_src = phase.state_options[name]['rate_source']
            if rate_src in phase.parameter_options:
                nn = num_seg * output_nodes_per_seg
                shape = phase.parameter_options[rate_src]['shape']
                param_size = np.prod(shape)
                src_idxs = np.tile(np.arange(0, param_size, dtype=int), nn)
                src_idxs = np.reshape(src_idxs, (nn,) + shape)
                phase.promotes('timeseries', inputs=[(f'all_values:state_rates:{name}',
                                                      f'parameters:{rate_src}')],
                               src_indices=src_idxs, src_shape=shape)
            else:
                phase.connect(src_name=self.get_rate_source_path(name, phase),
                              tgt_name=f'timeseries.all_values:state_rates:{name}')

        for name, options in phase.control_options.items():
            control_units = options['units']

            timeseries_comp._add_output_configure(f'controls:{name}',
                                                  shape=options['shape'],
                                                  units=control_units,
                                                  desc=options['desc'])

            phase.connect(src_name=f'control_values:{name}',
                          tgt_name=f'timeseries.all_values:controls:{name}')

            # Control rates
            timeseries_comp._add_output_configure(f'control_rates:{name}_rate',
                                                  shape=options['shape'],
                                                  units=get_rate_units(control_units, time_units,
                                                                       deriv=1),
                                                  desc=f'first time-derivative of control {name}')

            phase.connect(src_name=f'control_rates:{name}_rate',
                          tgt_name=f'timeseries.all_values:control_rates:{name}_rate')

            # Control second derivatives
            timeseries_comp._add_output_configure(f'control_rates:{name}_rate2',
                                                  shape=options['shape'],
                                                  units=get_rate_units(control_units, time_units,
                                                                       deriv=2),
                                                  desc=f'first time-derivative of control {name}')

            phase.connect(src_name=f'control_rates:{name}_rate2',
                          tgt_name=f'timeseries.all_values:control_rates:{name}_rate2')

        for name, options in phase.polynomial_control_options.items():
            control_units = options['units']
            timeseries_comp._add_output_configure(f'polynomial_controls:{name}',
                                                  shape=options['shape'],
                                                  units=control_units,
                                                  desc=options['desc'])

            phase.connect(src_name=f'polynomial_control_values:{name}',
                          tgt_name=f'timeseries.all_values:polynomial_controls:{name}')

            # Polynomial control rates
            timeseries_comp._add_output_configure(f'polynomial_control_rates:{name}_rate',
                                                  shape=options['shape'],
                                                  units=get_rate_units(control_units, time_units,
                                                                       deriv=1),
                                                  desc=f'first time-derivative of control {name}')

            phase.connect(src_name=f'polynomial_control_rates:{name}_rate',
                          tgt_name=f'timeseries.all_values:polynomial_control_rates:{name}_rate')

            # Polynomial control second derivatives
            timeseries_comp._add_output_configure(f'polynomial_control_rates:{name}_rate2',
                                                  shape=options['shape'],
                                                  units=get_rate_units(control_units, time_units,
                                                                       deriv=2),
                                                  desc=f'second time-derivative of control {name}')

            phase.connect(src_name=f'polynomial_control_rates:{name}_rate2',
                          tgt_name=f'timeseries.all_values:polynomial_control_rates:{name}_rate2')

        for name, options in phase.parameter_options.items():
            prom_name = f'parameters:{name}'

            if options['include_timeseries']:
                phase.timeseries._add_output_configure(prom_name,
                                                       desc='',
                                                       shape=options['shape'],
                                                       units=options['units'])

                if output_nodes_per_seg is None:
                    src_idxs_raw = np.zeros(self.grid_data.subset_num_nodes['all'], dtype=int)
                else:
                    src_idxs_raw = np.zeros(num_seg * output_nodes_per_seg, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, options['shape'])

                tgt_name = f'all_values:parameters:{name}'
                phase.promotes('timeseries', inputs=[(tgt_name, prom_name)],
                               src_indices=src_idxs, flat_src_indices=True)

        for var, options in phase._timeseries['timeseries']['outputs'].items():
            output_name = options['output_name']
            units = options.get('units', None)
            wildcard_units = options.get('wildcard_units', None)

            if '*' in var:  # match outputs from the ODE
                ode_outputs = {opts['prom_name']: opts for (k, opts) in
                               phase.ode.get_io_metadata(iotypes=('output',)).items()}
                matches = filter(list(ode_outputs.keys()), var)
            else:
                matches = [var]

            for v in matches:
                if '*' in var:
                    output_name = v.split('.')[-1]
                    units = ode_outputs[v]['units']
                    # check for wildcard_units override of ODE units
                    if v in wildcard_units:
                        units = wildcard_units[v]

                # Determine the path to the variable which we will be constraining
                # This is more complicated for path constraints since, for instance,
                # a single state variable has two sources which must be connected to
                # the path component.
                var_type = phase.classify_var(v)

                # Ignore any variables that we've already added (states, times, controls, etc)
                if var_type != 'ode':
                    continue

                # Skip the timeseries output if it does not appear to be shaped as a dynamic variable
                # If the full shape does not start with num_nodes, skip this variable.
                if self.is_static_ode_output(v, phase, self.num_output_nodes):
                    warnings.warn(f'Cannot add ODE output {v} to the timeseries output. It is '
                                  f'sized such that its first dimension != num_nodes.')
                    continue

                shape, units = get_source_metadata(phase.ode, src=v, user_shape=options['shape'],
                                                   user_units=units)

                try:
                    timeseries_comp._add_output_configure(output_name, shape=shape, units=units, desc='')
                except ValueError as e:  # OK if it already exists
                    if 'already exists' in str(e):
                        continue
                    else:
                        raise e

                # Failed to find variable, assume it is in the RHS
                phase.connect(src_name=f'ode.{v}',
                              tgt_name=f'timeseries.all_values:{output_name}')

    def get_parameter_connections(self, name, phase):
        """
        Returns info about a parameter's target connections in the phase.

        Parameters
        ----------
        name : str
            The name of the parameter for which connection information is desired.
        phase : dymos.Phase
            The phase object to which this transcription applies.

        Returns
        -------
        list of (paths, indices)
            A list containing a tuple of target paths and corresponding src_indices to which the
            given design variable is to be connected.
        """
        connection_info = []
        num_seg = self.grid_data.num_segments
        output_nodes_per_seg = self.options['output_nodes_per_seg']
        num_final_ode_nodes = self.grid_data.subset_num_nodes['all'] \
            if output_nodes_per_seg is None else num_seg * output_nodes_per_seg

        if name in phase.parameter_options:
            options = phase.parameter_options[name]
            ode_tgts = get_targets(ode=phase.ode, name=name, user_targets=options['targets'])

            dynamic = options['dynamic']
            shape = options['shape']

            # Connections to the final ODE
            if dynamic:
                src_idxs_raw = np.zeros(num_final_ode_nodes, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, shape)
                if shape == (1,):
                    src_idxs = src_idxs.ravel()
            else:
                src_idxs_raw = np.zeros(1, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, shape)
                src_idxs = np.squeeze(src_idxs, axis=0)

            connection_info.append(([f'ode.{tgt}' for tgt in ode_tgts], src_idxs))

        return connection_info

    def _get_boundary_constraint_src(self, var, loc):
        pass

    def get_rate_source_path(self, state_var, phase):
        """
        Return the rate source location for a given state name.

        Parameters
        ----------
        state_var : str
            Name of the state.
        phase : dymos.Phase
            Phase object containing the rate source.

        Returns
        -------
        str
            Path to the rate source.
        """
        var = phase.state_options[state_var]['rate_source']

        if var == 'time':
            rate_path = 'time'
        elif var == 'time_phase':
            rate_path = 'time_phase'
        elif phase.state_options is not None and var in phase.state_options:
            rate_path = f'state_mux_comp.states:{var}'
        elif phase.control_options is not None and var in phase.control_options:
            rate_path = f'control_values:{var}'
        elif phase.polynomial_control_options is not None and var in phase.polynomial_control_options:
            rate_path = f'polynomial_control_values:{var}'
        elif phase.parameter_options is not None and var in phase.parameter_options:
            rate_path = f'parameters:{var}'
        elif var.endswith('_rate') and phase.control_options is not None and \
                var[:-5] in phase.control_options:
            rate_path = f'control_rates:{var}'
        elif var.endswith('_rate2') and phase.control_options is not None and \
                var[:-6] in phase.control_options:
            rate_path = f'control_rates:{var}'
        elif var.endswith('_rate') and phase.polynomial_control_options is not None and \
                var[:-5] in phase.polynomial_control_options:
            rate_path = f'polynomial_control_rates:{var}'
        elif var.endswith('_rate2') and phase.polynomial_control_options is not None and \
                var[:-6] in phase.polynomial_control_options:
            rate_path = f'polynomial_control_rates:{var}'
        else:
            rate_path = f'ode.{var}'

        return rate_path



import openmdao.utils.hooks as hooks
import argparse
import sys
import os
from openmdao.utils.general_utils import warn_deprecation
from dymos.run_problem import modify_problem, run_problem
from unittest.mock import patch


def _simple_exec(script_name, user_args):
    """
    Use this as executor for commands that run as Problem commands.

    Parameters
    ----------
    script_name : string
        Name of the script to run.
    user_args: list of strings
        Any user supplied arguments for the script.
    """
    # use patch to avoid writing to sys.argv and sys.path
    with patch.object(sys, 'argv', [script_name] + user_args):
        with patch.object(sys, 'path', [os.path.dirname(script_name)] + sys.path[1:]):

            with open(script_name, 'rb') as fp:
                code = compile(fp.read(), script_name, 'exec')

            globals_dict = {
                '__file__': script_name,
                '__name__': '__main__',
                '__package__': None,
                '__cached__': None,
            }

            exec(code, globals_dict)

    return globals_dict


def dymos_cmd(argv=None):
    """
    The Dymos command line script.

    Parameters
    ----------
    argv : list
        User-declared arguments.

    Returns
    -------
    dict
        Dictionary of globals.
    """
    msg = """
    The Dymos command-line interface is deprecated and will be removed in dymos 1.0.0.
    This functionality is available by using the run_problem function from a python script instead.
    """
    warn_deprecation(msg)

    # pre-parse sys.argv to split between before and after '--'
    alt_args = argv if argv else sys.argv
    if '--' in alt_args:
        idx = alt_args.index('--')
        sys_args = alt_args[:idx]
        user_args = alt_args[idx + 1:]
        argv = sys_args  # don't change sys.argv
    else:
        user_args = []

    parser = argparse.ArgumentParser(description='Dymos Command Line Tool')
    parser.add_argument('script', type=str,
                        help='Python script that creates a Dymos problem to run')
    parser.add_argument('-s', '--simulate', action='store_true',
                        help='If given, perform simulation after solving the problem.')
    parser.add_argument('-n', '--no_solve', action='store_true',
                        help='If given, do not run the driver on the problem (simulate only)')
    parser.add_argument('-t', '--solution', default=None,
                        help='A previous solution record file (or explicit simulation record file)'
                             ' from which the initial guess should be loaded. (default: None)')
    parser.add_argument('-r', '--reset_grid', action='store_true',
                        help='If given, reset the grid to the specs given in the problem definition'
                             ' instead of the grid associated with the loaded solution.')
    parser.add_argument('-l', '--refine_limit', default=0,
                        help='The number of passes through the grid refinement algorithm'
                             ' to use. (default: 0)')
    parser.add_argument('-o', '--solution_record_file', default='dymos_solution.db',
                        help='Set the name of the case recorder file for solution results. '
                        '(default: dymos_solution.db)')
    parser.add_argument('-i', '--simulation_record_file', default='dymos_simulation.db',
                        help='Set the name of the case recorder file for simulation results. '
                        '(default: dymos_simulation.db)')
    parser.add_argument('-p', '--make_plots', action='store_true',
                        help='If given, automatically generate plots of all timeseries outputs.')
    parser.add_argument('-e', '--plot_dir', default='plots',
                        help='Set the name of the directory to store the timeseries plots. '
                        '(default: plots)')

    args = parser.parse_args(argv)  # sys.argv is used if argv parameter is None

    # make sure the loaded db is not being overwritten by the new db
    if args.solution == args.solution_record_file:
        db_copy = 'old_' + args.solution
        os.rename(args.solution, db_copy)
        args.solution = db_copy

    opts = {
        'refine_iteration_limit': int(args.refine_limit),
        'restart': args.solution,
        'simulate': args.simulate,
        'no_solve': args.no_solve,
        'reset_grid': args.reset_grid,
        'solution_record_file': args.solution_record_file,
        'simulation_record_file': args.simulation_record_file,
        'make_plots': args.make_plots,
        'plot_dir': args.plot_dir,
    }

    class DymosHooks:
        """
        DymosHooks is a lightweight class which provides the Dymos command-line OpenMDAO
        hooks with a state.

        Attributes
        ----------
        _pre_hooks_enabled : bool
            When True, the associated pre final_setup hooks will be allowed to run.
        _post_hooks_enabled : bool
            When True, the associated post final_setup hooks will be allowed to run.
        """

        def __init__(self):
            self._pre_hooks_enabled = True
            self._post_hooks_enabled = False

        def _pre_final_setup(self, prob):
            if not self._pre_hooks_enabled:
                return
            self._pre_hooks_enabled = False

            modify_problem(prob, restart=opts['restart'],
                           solution_record_file=opts['solution_record_file'])
            self._post_hooks_enabled = True

        def _post_final_setup(self, prob):
            if not self._post_hooks_enabled:
                return
            self._post_hooks_enabled = False

            refine_iterations = opts.get('refine_iteration_limit')
            run_problem(prob, refine_iteration_limit=refine_iterations,
                        run_driver=not opts['no_solve'],
                        simulate=opts['simulate'],
                        make_plots=opts['make_plots'],
                        solution_record_file=opts['solution_record_file'],
                        simulation_record_file=opts['simulation_record_file'],
                        plot_dir=opts['plot_dir']
                        )

    dymos_hooks = DymosHooks()

    hooks.use_hooks = True
    hooks._register_hook('final_setup',
                         'Problem',
                         pre=dymos_hooks._pre_final_setup,
                         post=dymos_hooks._post_final_setup)

    globals_dict = _simple_exec(args.script, user_args)  # run the script

    # suppress printing return value when running from command line
    if not sys.argv[0].endswith('dymos'):
        return globals_dict  # return globals for possible checking in unit tests


if __name__ == '__main__':
    dymos_cmd()



import openmdao.api as om

from ....utils.rk_methods import rk_methods
from ....utils.introspection import get_targets

from .runge_kutta_state_predict_comp import RungeKuttaStatePredictComp
from .runge_kutta_k_comp import RungeKuttaKComp


class RungeKuttaKIterGroup(om.Group):
    """
    Class definition for the RungeKuttaKIterGroup.

    This Group contains the state prediction component, the ODE, and the k-calculation component.
    Given the initial values of the states, the times at the ODE evaluations, and the stepsize
    across each segment it will iterate to find the Runge-Kutta weights 'k'.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('num_segments', types=int,
                             desc='The number of segments (timesteps) in the phase')

        self.options.declare('method', default='RK4', types=str,
                             desc='Specific Runge-Kutta Method to use.')

        self.options.declare('state_options', types=dict,
                             desc='Dictionary of state names/options for the phase')

        self.options.declare('time_units', default=None, allow_none=True, types=str,
                             desc='Units of the integration variable')

        self.options.declare('ode_class',
                             desc='System defining the ODE')

        self.options.declare('ode_init_kwargs', types=dict, default={},
                             desc='Keyword arguments provided when initializing the ODE System')

        self.options.declare('solver_class', default=om.NonlinearBlockGS,
                             values=(om.NonlinearBlockGS, om.NewtonSolver, om.NonlinearRunOnce),
                             allow_none=True,
                             desc='The nonlinear solver class used to converge the numerical '
                                  'integration of the segment.')

        self.options.declare('solver_options', default={}, types=(dict,),
                             desc='The options passed to the nonlinear solver used to converge the'
                                  'Runge-Kutta propagation across each step.')

    def setup(self):
        """
        Build the group hierarchy.
        """
        num_seg = self.options['num_segments']
        rk_data = rk_methods[self.options['method']]
        num_nodes = num_seg * rk_data['num_stages']
        state_options = self.options['state_options']

        self.add_subsystem('state_predict_comp',
                           RungeKuttaStatePredictComp(method=self.options['method'],
                                                      num_segments=num_seg,
                                                      state_options=state_options))
        self.add_subsystem('ode',
                           subsys=self.options['ode_class'](num_nodes=num_nodes,
                                                            **self.options['ode_init_kwargs']))

        self.add_subsystem('k_comp',
                           subsys=RungeKuttaKComp(method=self.options['method'],
                                                  num_segments=num_seg,
                                                  state_options=state_options,
                                                  time_units=self.options['time_units']))

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine variable shape and units.
        """
        self.state_predict_comp.configure_io()
        self.k_comp.configure_io()

        for state_name, options in self.options['state_options'].items():
            targets = get_targets(ode=self.ode, name=state_name, user_targets=options['targets'])

            # Connect the state predicted (assumed) value to its targets in the ode
            if targets:
                self.connect('state_predict_comp.predicted_states:{0}'.format(state_name),
                             ['ode.{0}'.format(tgt) for tgt in targets])

            # Connect the k value associated with the state to the state predict comp
            self.connect('k_comp.k:{0}'.format(state_name),
                         'state_predict_comp.k:{0}'.format(state_name))

            # Delayed promotes
            self.promotes('state_predict_comp', inputs=[f'initial_states_per_seg:{state_name}'])
            self.promotes('k_comp', inputs=['h'])

        self.linear_solver = om.DirectSolver()
        if self.options['solver_class']:
            self.nonlinear_solver = self.options['solver_class'](**self.options['solver_options'])



import numpy as np


def get_src_indices_by_row(row_idxs, shape, flat=True):
    """
    Provide the src_indices when connecting a vectorized variable from an output to an input.

    Indices are selected by choosing the first indices to be passed, corresponding to node
    index in Dymos.

    Parameters
    ----------
    row_idxs : array_like
        The rows/node indices to be connected from the source to the target.
    shape : tuple
        The shape of the variable at each node (ignores the first dimension).
    flat : bool
        If True, return the source indices in flat source indices form.

    Returns
    -------
    array_like
        If flat, a numpy array of shape `(row_idxs,) + shape` where each element is the index
        of the source of that element in the source array, in C-order.
    """
    if not flat:
        raise NotImplementedError('Currently get_src_indices_by_row only returns '
                                  'flat source indices.')

    num_src_rows = np.max(row_idxs) + 1
    src_shape = (num_src_rows,) + shape
    other_idxs = [np.arange(n, dtype=int) for n in shape]
    ixgrid = np.ix_(row_idxs, *other_idxs)
    a = np.reshape(np.arange(np.prod(src_shape), dtype=int), newshape=src_shape)
    src_idxs = a[ixgrid]
    return src_idxs



import numpy as np
from scipy.linalg import block_diag

import openmdao.api as om

from ....utils.rk_methods import rk_methods
from ....options import options as dymos_options


class RungeKuttaStatePredictComp(om.ExplicitComponent):
    """
    Class definition for the RungeKuttaStatePredictComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('num_segments', types=int,
                             desc='The number of segments (timesteps) in the phase')

        self.options.declare('method', default='RK4', types=str,
                             desc='Specific Runge-Kutta Method to use.')

        self.options.declare('state_options', types=dict,
                             desc='Dictionary of state names/options for the phase')

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine variable shape and units.
        """
        self._var_names = {}
        num_seg = self.options['num_segments']

        rk_data = rk_methods[self.options['method']]
        self._A = block_diag(rk_data['A'])
        self._num_stages = rk_data['num_stages']

        for name, options in self.options['state_options'].items():
            shape = options['shape']
            units = options['units']

            self._var_names[name] = {}
            self._var_names[name]['initial'] = 'initial_states_per_seg:{0}'.format(name)
            self._var_names[name]['k'] = 'k:{0}'.format(name)
            self._var_names[name]['predicted'] = 'predicted_states:{0}'.format(name)

            self.add_input(self._var_names[name]['initial'], shape=(num_seg,) + shape, units=units,
                           desc='The initial value of the state at the start of the segment.')

            self.add_input(self._var_names[name]['k'],
                           shape=(num_seg, self._num_stages,) + shape,
                           units=units, desc='RK multiplier k for each stage in the segment.')

            self.add_output(self._var_names[name]['predicted'],
                            shape=(num_seg * self._num_stages,) + shape,
                            units=units,
                            desc='The predicted values of the state at the ODE evaluation points.')

            e = np.eye(np.prod(shape))
            p = np.kron(np.ones(self._num_stages), e).T
            p = block_diag(*num_seg*[p])
            r, c = np.nonzero(p)
            self.declare_partials(of=self._var_names[name]['predicted'],
                                  wrt=self._var_names[name]['initial'],
                                  rows=r, cols=c, val=1.0)

            size = np.prod(shape)
            p = block_diag(*num_seg*[np.kron(self._A, np.eye(size))])
            r, c = np.nonzero(p)
            self.declare_partials(of=self._var_names[name]['predicted'],
                                  wrt=self._var_names[name]['k'],
                                  rows=r, cols=c, val=p[r, c])

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        #
        # Note:  To accommodate states of dimension 2 or more, k is flattened to
        # (num_segments * num_stage, size) prior to the matrix vector product,
        # and then reshaped back to the appropriate one.  This is complicated by the fact that
        # we need k in (num_stages, num_segments, size) for the matrix multiply to work out.
        # See below for the steps taken.
        #
        num_seg = self.options['num_segments']
        num_stages = self._num_stages
        for name, options in self.options['state_options'].items():
            shape = options['shape']
            size = np.prod(shape)
            x0 = inputs[self._var_names[name]['initial']]

            # # Reorder k to (num_stages, num_segments, size) to avoid a for loop
            # # k = np.moveaxis(inputs[self._var_names[name]['k']],
            # #                 [0, 1], [1, 0]).reshape((num_stages, size * num_seg))
            # # Einsum appears to be faster than moveaxis, so we use this implementation instead.
            k = np.einsum('ij...->ji...',
                          inputs[self._var_names[name]['k']]).reshape((num_stages, size * num_seg))

            # The matrix multiply
            A_dot_k = np.dot(self._A, k)

            # Now stack the segments vertically so we have num_seg * num_stages rows and size cols
            A_dot_k = np.vstack(np.split(A_dot_k, num_seg, axis=1))

            # Now give x0 the same treatment.
            # First broadcast to num_stages rows and num_segments * size cols
            _x0 = np.repeat(np.reshape(x0, (1, num_seg * size)), num_stages, axis=0)

            # Now chunk x0 into num_segments * num_stages rows and size cols
            _x0 = np.vstack(np.split(_x0, num_seg, axis=1))

            # Now compute the predicted state and change back to the correct shape at each row
            x_p = np.reshape(_x0 + A_dot_k, newshape=(num_seg * num_stages,) + shape)

            outputs[self._var_names[name]['predicted']] = x_p



import numpy as np


class _ReprClass(object):
    """
    Class for defining objects with a simple constant string __repr__.

    This is useful for constants used in arg lists when you want them to appear in
    automatically generated source documentation as a certain string instead of python's
    default representation.
    """

    def __init__(self, repr_string):
        """
        Inititialize the __repr__ string.

        Parameters
        ----------
        repr_string : str
            The string to be returned by __repr__
        """
        self._repr_string = repr_string

    def __repr__(self):
        """
        Return our _repr_string.

        Returns
        -------
        str
            Whatever string we were initialized with.
        """
        return self._repr_string

    def __str__(self):
        return self._repr_string


# unique object to check if default is given (when None is an allowed value)
_unspecified = _ReprClass("unspecified")


def get_rate_units(units, time_units, deriv=1):
    """
    Return a string for rate units given units for the variable and time units.

    Parameters
    ----------
    units : str
        Units of a given variable.
    time_units : str
        Time units.
    deriv : int
        If 1, provide the units of the first derivative.  If 2,
        provide the units of the second derivative.

    Returns
    -------
    str
        Corresponding rate units for the given variable.
    """
    if deriv not in (1, 2):
        raise ValueError('deriv argument must be 1 or 2.')

    tu = time_units if deriv == 1 else '{0}**2'.format(time_units)

    if units is not None and time_units is not None:
        rate_units = '{0}/{1}'.format(units, tu)
    elif units is not None:
        rate_units = units
    elif time_units is not None:
        rate_units = '1.0/{0}'.format(tu)
    else:
        rate_units = None
    return rate_units


def get_target_metadata(ode, name, user_targets=_unspecified, user_units=_unspecified,
                        user_shape=_unspecified, control_rate=False, dynamic=True):
    """
    Return the targets of a state variable in a given ODE system.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : om.System
        The OpenMDAO system which serves as the ODE for dymos.  This system should already have
        had its setup and configure methods called.
    name : str
        The name of the state variable whose targets are desired.
    user_targets : str or None or Sequence or _unspecified
        Targets for the variable as given by the user.
    user_units : str or None or Sequence or _unspecified
        Units for the variable as given by the user.
    user_shape : str or None or Sequence or _unspecified
        Shape for the variable as given by the user.
    control_rate : bool
        When True, check for the control rate if the name is not in the ODE.
    dynamic : bool
        When True, assume the shape of the target in the ODE includes the number of nodes as the
        first dimension.

    Returns
    -------
    shape : tuple
        The shape of the variable.  If not specified, shape is taken from the ODE targets.
    units : str
        The units of the variable.  If not specified, units are taken from the ODE targets.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    rate_src = False
    ode_inputs = {opts['prom_name']: opts for (k, opts) in
                  ode.get_io_metadata(iotypes=('input', 'output'), get_remote=True).items()}

    if user_targets is _unspecified:
        if name in ode_inputs:
            targets = [name]
        elif control_rate and f'{name}_rate' in ode_inputs:
            targets = [f'{name}_rate']
            rate_src = True
        else:
            targets = []
    elif user_targets:
        if isinstance(user_targets, str):
            targets = [user_targets]
        else:
            targets = user_targets
    else:
        targets = []

    if user_units is _unspecified:
        target_units_set = {ode_inputs[tgt]['units'] for tgt in targets}
        if len(target_units_set) == 1:
            units = next(iter(target_units_set))
            if rate_src:
                units = f"{units}*s"
        else:
            raise ValueError(f'Unable to automatically assign units to {name}. '
                             f'Targets have multiple units: {target_units_set}. '
                             f'Either promote targets and use set_input_defaults to assign common '
                             f'units, or explicitly provide them to {name}.')
    else:
        units = user_units

    if user_shape in {None, _unspecified}:
        target_shape_set = {ode_inputs[tgt]['shape'] for tgt in targets}
        if len(target_shape_set) == 1:
            shape = next(iter(target_shape_set))
            if dynamic:
                if len(shape) == 1:
                    shape = (1,)
                else:
                    shape = shape[1:]
        elif len(target_shape_set) == 0:
            raise ValueError(f'Unable to automatically assign a shape to {name}.\n'
                             'Targets for this variable either do not exist or have no shape set.\n'
                             'The shape for this variable must be set explicitly via the '
                             '`shape=<tuple>` argument.')
        else:
            raise ValueError(f'Unable to automatically assign a shape to {name} based on targets. '
                             f'Targets have multiple shapes assigned: {target_shape_set}. '
                             f'Change targets such that all have common shapes.')
    else:
        shape = user_shape

    return shape, units


def get_source_metadata(ode, src, user_units, user_shape):
    """
    Return the targets of a state variable in a given ODE system.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : om.System
        The OpenMDAO system which serves as the ODE for dymos.  This system should already have
        had its setup and configure methods called.
    src : str
        The relative path in the ODE to the source variable whose metadata is requested.
    user_units : str or None or Sequence or _unspecified
        Units for the variable as given by the user.
    user_shape : str or None or Sequence or _unspecified
        Shape for the variable as given by the user.

    Returns
    -------
    shape : tuple
        The shape of the variable.  If not specified, shape is taken from the ODE targets.
    units : str
        The units of the variable.  If not specified, units are taken from the ODE targets.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    ode_outputs = {opts['prom_name']: opts for (k, opts) in
                   ode.get_io_metadata(iotypes=('output',), get_remote=True).items()}

    if src not in ode_outputs:
        raise ValueError(f'Unable to find the source {src} in the ODE at {ode.pathname}.')

    if user_units in {None, _unspecified}:
        units = ode_outputs[src]['units']
    else:
        units = user_units

    if user_shape in {None, _unspecified}:
        ode_shape = ode_outputs[src]['shape']
        shape = (1,) if len(ode_shape) == 1 else ode_shape[1:]
    else:
        shape = user_shape

    return shape, units


def reshape_val(val, shape, num_input_nodes):
    """
    Return the given value reshaped to (num_input_nodes,) + shape.

    If the value is scalar or a size-1 array, return that value multiplied by
    np.ones((num_input_nodes,) + shape).  If the value's shape is shape, then
    repeat those values along a new first dimension.  Otherwise, reshape it to
    the correct shape and return that.

    Parameters
    ----------
    val : float or array-like
        The values to be conformed to the desired shape.
    shape : tuple
        The shape of the desired output at each node.
    num_input_nodes : int
        The number of nodes along which the value is repeated.

    Returns
    -------
    np.array
        The given value of the correct shape.
    """
    if np.isscalar(val) or np.prod(np.asarray(val).shape) == 1:
        shaped_val = float(val) * np.ones((num_input_nodes,) + shape)
    elif np.asarray(val).shape == shape:
        shaped_val = np.repeat(val[np.newaxis, ...], num_input_nodes, axis=0)
    else:
        shaped_val = np.reshape(val, newshape=(num_input_nodes,) + shape)
    return shaped_val


class CoerceDesvar(object):
    """
    Check the desvar options for the appropriate shape and resize accordingly with options.

    Parameters
    ----------
    num_input_nodes : int
        Number of input nodes.
    desvar_indices : ndarray
        Flattened indices of the variable.
    options : dict
        Variable options dictionary, should contain "shape".
    """
    def __init__(self, num_input_nodes, desvar_indices, options):
        self.num_input_nodes = num_input_nodes
        self.desvar_indices = desvar_indices
        self.options = options

    def __call__(self, option):
        """
        Test that an option's shape is compliant with the number of input nodes for the design variable.

        Parameters
        ----------
        option : str
            The name of the option whose value(s) are desired.

        Returns
        -------
        object
            The value of the desvar option.

        Raises
        ------
        ValueError
            If the number of values in the option is not compliant with the number of input
            nodes for the design variable.

        """
        val = self.options[option]
        if val is None or np.isscalar(val):
            return val
        # Handle value for vector/matrix valued variables
        if isinstance(val, list):
            val = np.asarray(val)
        if val.shape == self.options['shape']:
            return np.tile(val.flatten(), int(len(self.desvar_indices)/val.size))
        else:
            raise ValueError('array-valued option {0} must have same shape '
                             'as states ({1})'.format(option, self.options['shape']))


def CompWrapperConfig(comp_class):
    """
    Returns a wrapped comp_class that calls its configure_io method at the end of setup.

    This allows for standalone testing of Dymos components that normally require their parent group
    to configure them.

    Parameters
    ----------
    comp_class : Component class
       Class that we would like to wrap.

    Returns
    -------
    WrappedClass
        Wrapped version of comp_class.
    """
    class WrappedClass(comp_class):

        def setup(self):
            """
            Appends a call to configure_io after setup.
            """
            super(WrappedClass, self).setup()
            self.configure_io()

    return WrappedClass


# Modify class so we can run it standalone.
def GroupWrapperConfig(comp_class):
    """
    Returns a wrapped group_class that calls its configure_io method at the end of setup.

    This allows for standalone testing of Dymos components that normally require their parent group
    to configure them.

    Parameters
    ----------
    comp_class : Group class
       Class that we would like to wrap.

    Returns
    -------
    WrappedClass
        Wrapped version of comp_class.
    """
    class WrappedClass(comp_class):

        def setup(self):
            """
            Setup as normal.
            """
            super(WrappedClass, self).setup()

        def configure(self):
            """
            Call configure_io during configure.
            """
            self.configure_io()

    return WrappedClass



import numpy as np

import openmdao.api as om

from ...utils.constants import INF_BOUND
from ...options import options as dymos_options


class BoundaryConstraintComp(om.ExplicitComponent):
    """
    Class definition for the BoundaryConstraintComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('loc', values=('initial', 'final'),
                             desc='the location in the phase of this boundary constraint '
                                  '(either \'initial\' or \'final\'')
        self._constraints = []
        self._vars = {}

    def configure_io(self):
        """
        Define the independent variables as output variables.

        I/O creation is delayed until configure so that we can determine the shape and units for
        the states.
        """
        for (name, kwargs) in self._constraints:
            input_name = '{0}_value_in:{1}'.format(self.options['loc'], name)
            output_name = '{0}_value:{1}'.format(self.options['loc'], name)
            self._vars[name] = {'input_name': input_name,
                                'output_name': output_name,
                                'shape': kwargs['shape']}

            input_kwargs = {k: kwargs[k] for k in ('units', 'shape', 'desc')}
            self.add_input(input_name, **input_kwargs)

            output_kwargs = {k: kwargs[k] for k in ('units', 'shape', 'desc')}
            self.add_output(output_name, **output_kwargs)

            constraint_kwargs = {k: kwargs.get(k, None)
                                 for k in ('lower', 'upper', 'equals', 'ref', 'ref0', 'adder',
                                           'scaler', 'indices', 'linear')}
            self.add_constraint(output_name, **constraint_kwargs)

        # Setup partials
        for name, options in self._vars.items():
            size = int(np.prod(options['shape']))

            rs = np.arange(size)
            cs = np.arange(size)

            self.declare_partials(of=options['output_name'],
                                  wrt=options['input_name'],
                                  val=np.ones(size),
                                  rows=rs,
                                  cols=cs)

    def compute(self, inputs, outputs):
        """
        Compute boundary constraint outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        for name, options in self._vars.items():
            outputs[options['output_name']] = inputs[options['input_name']]

    def _add_constraint(self, name, units=None, res_units=None, desc='',
                        shape=None, indices=None, flat_indices=True,
                        lower=None, upper=None, equals=None,
                        scaler=None, adder=None, ref=1.0, ref0=0.0,
                        linear=False, res_ref=1.0, distributed=False):
        """
        Add a constraint to this component

        Parameters
        ----------
        name : str
            name of the variable in this component's namespace.
        val : float or list or tuple or ndarray
            The initial value of the variable being added in user-defined units. Default is 1.0.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        indices : tuple, list, ndarray, or None
            The indices of the output variable to be boundary constrained.  If provided, the
            resulting constraint is always a 1D vector with the number of elements provided in
            indices.  Indices should be a 1D sequence of tuples, each providing an index into the
            source output if flat_indices is False, or integers if flat_indices is True.
        flat_indices : bool
            Whether or not indices is provided as 'flat' indices per OpenMDAO's flat_source_indices
            option when connecting variables.
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        desc : str
            description of the variable
        lower : float or list or tuple or ndarray or None
            lower bound(s) in user-defined units. It can be (1) a float, (2) an array_like
            consistent with the shape arg (if given), or (3) an array_like matching the shape of
            val, if val is array_like. A value of None means this output has no lower bound.
            Default is None.
        upper : float or list or tuple or ndarray or None
            upper bound(s) in user-defined units. It can be (1) a float, (2) an array_like
            consistent with the shape arg (if given), or (3) an array_like matching the shape of
            val, if val is array_like. A value of None means this output has no upper bound.
            Default is None.
        scaler : float or None
            A multiplicative scaler on the constraint value for the optimizer.
        adder : float or None
            A parameter which is added to the value before scaler is applied to produce
            the value seen by the optimizer.
        ref : float or None
            Scaling parameter. The value in the user-defined units of this output variable when
            the scaled value is 1. Default is 1.
        ref0 : float or None
            Scaling parameter. The value in the user-defined units of this output variable when
            the scaled value is 0. Default is 0.
        linear : bool
            True if the *total* derivative of the constrained variable is linear, otherwise False.
        distributed : bool
            If True, this variable is distributed across multiple processes.
        """
        lower = -INF_BOUND if upper is not None and lower is None else lower
        upper = INF_BOUND if lower is not None and upper is None else upper
        kwargs = {'units': units, 'res_units': res_units, 'desc': desc,
                  'shape': shape, 'indices': indices, 'flat_indices': flat_indices,
                  'lower': lower, 'upper': upper, 'equals': equals,
                  'scaler': scaler, 'adder': adder, 'ref': ref, 'ref0': ref0, 'linear': linear,
                  'res_ref': res_ref, 'distributed': distributed}
        self._constraints.append((name, kwargs))



import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt


def vanderpol_dymos_plots(p):
    """ Plot results of a given Dymos VanDerPol ODE problem """

    # check validity by using scipy.integrate.solve_ivp to integrate the solution
    sim_out = p.model.traj.simulate()

    # plots
    fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12, 9))

    # state variable x1
    axes[0, 0].plot(p.get_val('traj.phase0.timeseries.time'),
                    p.get_val('traj.phase0.timeseries.states:x1'),
                    'ro', label='solution')

    axes[0, 0].plot(sim_out.get_val('traj.phase0.timeseries.time'),
                    sim_out.get_val('traj.phase0.timeseries.states:x1'),
                    'b-', label='simulation')

    axes[0, 0].set_xlabel('time (s)')
    axes[0, 0].set_ylabel('x1 (V)')
    axes[0, 0].legend()
    axes[0, 0].grid()

    # state variable x0
    axes[0, 1].plot(p.get_val('traj.phase0.timeseries.time'),
                    p.get_val('traj.phase0.timeseries.states:x0'),
                    'ro', label='solution')

    axes[0, 1].plot(sim_out.get_val('traj.phase0.timeseries.time'),
                    sim_out.get_val('traj.phase0.timeseries.states:x0'),
                    'b-', label='simulation')

    axes[0, 1].set_xlabel('time (s)')
    axes[0, 1].set_ylabel('x0 (V/s)')
    axes[0, 1].legend()
    axes[0, 1].grid()

    # state variable x1 vs x0
    axes[1, 0].plot(p.get_val('traj.phase0.timeseries.states:x0'),
                    p.get_val('traj.phase0.timeseries.states:x1'),
                    'ro', label='solution')

    axes[1, 0].plot(sim_out.get_val('traj.phase0.timeseries.states:x0'),
                    sim_out.get_val('traj.phase0.timeseries.states:x1'),
                    'b-', label='simulation')

    axes[1, 0].set_xlabel('x0 (V/s)')
    axes[1, 0].set_ylabel('x1 (V)')
    axes[1, 0].legend()
    axes[1, 0].grid()

    # control variable u
    axes[1, 1].plot(p.get_val('traj.phase0.timeseries.time'),
                    p.get_val('traj.phase0.timeseries.controls:u'),
                    'ro', label='solution')

    axes[1, 1].plot(sim_out.get_val('traj.phase0.timeseries.time'),
                    sim_out.get_val('traj.phase0.timeseries.controls:u'),
                    'b-', label='simulation')

    axes[1, 1].set_xlabel('time (s)')
    axes[1, 1].set_ylabel('control u')
    axes[1, 1].legend()
    axes[1, 1].grid()

    plt.show()



import warnings

from .grid_refinement.ph_adaptive.ph_adaptive import PHAdaptive
from .grid_refinement.hp_adaptive.hp_adaptive import HPAdaptive
from .grid_refinement.write_iteration import write_error, write_refine_iter
from .grid_refinement.refinement import _refine_iter
from .phase.phase import Phase

import openmdao.api as om
import dymos as dm
import numpy as np
from dymos.trajectory.trajectory import Trajectory
from dymos.load_case import load_case, find_phases
from dymos.visualization.timeseries_plots import timeseries_plots
from dymos.grid_refinement.error_estimation import check_error
import os
import sys


def modify_problem(problem, restart=None, reset_grid=False,
                   solution_record_file='dymos_solution.db'):
    """
    Modifies the problem object by loading in a guess from a specified restart file.
    Parameters
    ----------
    problem : om.Problem
        The problem instance being modified.
    restart : String or None
        The name of a database to use for restarting the problem.
    reset_grid: Boolean
        Flag to trigger a grid reset. Not implemented yet.
    solution_record_file: str
        The path to the file to be used for recording the solution recording.
    """
    # record variables to database when running driver under hook
    # pre-hook is important, because recording initialization is skipped if final_setup has run once
    try:
        os.remove(solution_record_file)
    except FileNotFoundError:
        pass  # OK if old database is not present to be deleted

    print('adding recorder at:', solution_record_file)
    problem.add_recorder(om.SqliteRecorder(solution_record_file))

    # if opts.get('reset_grid'):  # TODO: implement this option
    #     pass

    if restart is not None:  # restore variables from database file specified by 'restart'
        print('Restarting run_problem using the %s database.' % restart)
        cr = om.CaseReader(restart)

        # find the proper case
        try:
            case = cr.get_case('final')
        except RuntimeError:
            cases = cr.list_cases()
            if len(cases) < 1:
                print('WARNING: the requested %s database file does not have any cases to load.'
                      % restart)
                return
            case = cr.get_case(cases[-1])  # use last case, ideally it should be the only one

        check_simulation = cr.problem_metadata['driver']['name'] == 'Driver'
        if check_simulation:
            prev_soln = {'inputs':  case.list_inputs(out_stream=None,  units=True, prom_name=True),
                         'outputs': case.list_outputs(out_stream=None, units=True, prom_name=True)}

            load_case(problem, prev_soln)
        else:
            # Initialize the system with values from the case.
            # We unnecessarily call setup again to make sure we obliterate the previous solution
            # First reset the connections at the top level model until fixed in OpenMDAO
            problem.setup()

            # Load the values from the previous solution
            load_case(problem, case)


def run_problem(problem, refine_method='hp', refine_iteration_limit=0, run_driver=True,
                simulate=False, restart=None,
                solution_record_file='dymos_solution.db',
                simulation_record_file='dymos_simulation.db',
                make_plots=False,
                plot_dir="plots"
                ):
    """
    A Dymos-specific interface to execute an OpenMDAO problem containing Dymos Trajectories or
    Phases.  This function can iteratively call run_driver to perform grid refinement, and
    automatically call simulate following a run to check the validity of a result.

    Parameters
    ----------
    problem : om.Problem
        The OpenMDAO problem object to be run.
    refine_method : String
        The choice of refinement algorithm to use for grid refinement
    refine_iteration_limit : int
        The number of passes through the grid refinement algorithm to be made.
    run_driver : bool
        If True, run the driver (optimize the problem), otherwise just run the model one time.
    simulate : bool
        If True, perform a simulation of Trajectories found in the Problem after the driver
        has been run and grid refinement is complete.
    restart : str
        If not None, the path to a CaseRecorder file used to load in recording from a previous run.
    make_plots : bool
        If True, automatically generate plots of all timeseries outputs.
    solution_record_file : String
        Path to case recorder file use to store results from solution.
    simulation_record_file : String
        Path to case recorder file use to store results from simulation.
    plot_dir : str
        Path to directory for plot files.
    """
    if restart is not None:
        case = om.CaseReader(restart).get_case('final')

    if solution_record_file not in [rec._filepath for rec in iter(problem._rec_mgr)]:
        recorder = om.SqliteRecorder(solution_record_file)
        problem.add_recorder(recorder)
        # record_inputs is needed to capture potential input parameters that aren't connected
        problem.recording_options['record_inputs'] = True
        # record_outputs is need to capture the timeseries outputs
        problem.recording_options['record_outputs'] = True

    problem.final_setup()  # make sure command line option hook has a chance to run

    if restart is not None:
        load_case(problem, case)

    if run_driver:
        problem.run_driver()
        _refine_iter(problem, refine_iteration_limit, refine_method)
    else:
        problem.run_model()
        if refine_iteration_limit > 0:
            warnings.warn("Refinement not performed. Set run_driver to True to perform refinement.")

    problem.record('final')  # save case for potential restart
    problem.cleanup()

    if simulate:
        for subsys in problem.model.system_iter(include_self=True, recurse=True):
            if isinstance(subsys, Trajectory):
                subsys.simulate(record_file=simulation_record_file)

    if make_plots:
        if simulate:
            timeseries_plots(solution_record_file, simulation_record_file=simulation_record_file,
                             plot_dir=plot_dir)
        else:
            timeseries_plots(solution_record_file, simulation_record_file=None,
                             plot_dir=plot_dir)
